@@ -17,13 +17,18 @@
  *     Ted Gould <ted.gould@canonical.com>
  */
 
+#include <future>
+#include <thread>
+
 #include <gtest/gtest.h>
 #include <gio/gio.h>
 #include <zeitgeist.h>
+#include "mir-mock.h"
 
 extern "C" {
 #include "ubuntu-app-launch.h"
 #include "libdbustest/dbus-test.h"
+#include <fcntl.h>
 }
 
 class LibUAL : public ::testing::Test
@@ -79,7 +84,8 @@ class LibUAL : public ::testing::Test
 			g_free(linkfarmpath);
 
 			g_setenv("XDG_DATA_DIRS", CMAKE_SOURCE_DIR, TRUE);
-			g_setenv("XDG_CACHE_HOME", CMAKE_SOURCE_DIR, TRUE);
+			g_setenv("XDG_CACHE_HOME", CMAKE_SOURCE_DIR "/libertine-data", TRUE);
+			g_setenv("XDG_DATA_HOME",  CMAKE_SOURCE_DIR "/libertine-home", TRUE);
 
 			service = dbus_test_service_new(NULL);
 
@@ -108,6 +114,13 @@ class LibUAL : public ::testing::Test
 				"	ret = dbus.ObjectPath('/com/test/untrusted/helper')\n",
 				NULL);
 
+			dbus_test_dbus_mock_object_add_method(mock, obj,
+				"SetEnv",
+				G_VARIANT_TYPE("(assb)"),
+				NULL,
+				"",
+				NULL);
+
 			/* Click App */
 			DbusTestDbusMockObject * jobobj = dbus_test_dbus_mock_get_object(mock, "/com/test/application_click", "com.ubuntu.Upstart0_6.Job", NULL);
 
@@ -266,50 +279,64 @@ class LibUAL : public ::testing::Test
 			}
 			ASSERT_EQ(nullptr, bus);
 		}
-
-		bool check_env (GVariant * env_array, const gchar * var, const gchar * value) {
-			GVariantIter iter;
-			g_variant_iter_init(&iter, env_array);
+		
+		GVariant * find_env (GVariant * env_array, const gchar * var) {
+			int i;
 			gchar * envvar = NULL;
-			bool found = false;
+			GVariant * retval = nullptr;
+
+			for (i = 0; i < g_variant_n_children(env_array); i++) {
+				GVariant * child = g_variant_get_child_value(env_array, i);
+				const gchar * envvar = g_variant_get_string(child, nullptr);
 
-			while (g_variant_iter_loop(&iter, "s", &envvar)) {
 				if (g_str_has_prefix(envvar, var)) {
-					if (found) {
+					if (retval != nullptr) {
 						g_warning("Found the env var more than once!");
-						return false;
+						g_variant_unref(retval);
+						return nullptr;
 					}
 
-					if (value != NULL) {
-						gchar * combined = g_strdup_printf("%s=%s", var, value);
-						if (g_strcmp0(envvar, combined) == 0) {
-							found = true;
-						}
-						g_free(combined);
-					} else {
-						found = true;
-					}
+					retval = child;
+				} else {
+					g_variant_unref(child);
 				}
 			}
 
-			if (!found) {
+			if (!retval) {
 				gchar * envstr = g_variant_print(env_array, FALSE);
-				g_warning("Unable to find '%s' with value '%s' in '%s'", var, value, envstr);
+				g_warning("Unable to find '%s' in '%s'", var, envstr);
 				g_free(envstr);
 			}
 
+			return retval;
+		}
+
+		bool check_env (GVariant * env_array, const gchar * var, const gchar * value) {
+			bool found = false;
+			GVariant * val = find_env(env_array, var);
+			if (val == nullptr)
+				return false;
+
+			const gchar * envvar = g_variant_get_string(val, nullptr);
+
+			gchar * combined = g_strdup_printf("%s=%s", var, value);
+			if (g_strcmp0(envvar, combined) == 0) {
+				found = true;
+			}
+
+			g_variant_unref(val);
+
 			return found;
 		}
 
-		static gboolean pause_helper (gpointer pmainloop) {
-			g_main_loop_quit(static_cast<GMainLoop *>(pmainloop));
-			return G_SOURCE_REMOVE;
-		}
-
-		void pause (guint time) {
+		void pause (guint time = 0) {
 			if (time > 0) {
 				GMainLoop * mainloop = g_main_loop_new(NULL, FALSE);
-				g_timeout_add(time, pause_helper, mainloop);
+
+				g_timeout_add(time, [](gpointer pmainloop) -> gboolean {
+					g_main_loop_quit(static_cast<GMainLoop *>(pmainloop));
+					return G_SOURCE_REMOVE;
+				}, mainloop);
 
 				g_main_loop_run(mainloop);
 
@@ -409,18 +436,22 @@ TEST_F(LibUAL, StopApplication)
 
 }
 
+/* NOTE: The fact that there is 'libertine-data' in these strings is because
+   we're using one CACHE_HOME for this test suite and the libertine functions
+   need to pull things from there, where these are only comparisons. It's just
+   what value is in the environment variable */
 TEST_F(LibUAL, ApplicationLog)
 {
 	gchar * click_log = ubuntu_app_launch_application_log_path("com.test.good_application_1.2.3");
-	EXPECT_STREQ(CMAKE_SOURCE_DIR "/upstart/application-click-com.test.good_application_1.2.3.log", click_log);
+	EXPECT_STREQ(CMAKE_SOURCE_DIR "/libertine-data/upstart/application-click-com.test.good_application_1.2.3.log", click_log);
 	g_free(click_log);
 
 	gchar * legacy_single = ubuntu_app_launch_application_log_path("single");
-	EXPECT_STREQ(CMAKE_SOURCE_DIR "/upstart/application-legacy-single-.log", legacy_single);
+	EXPECT_STREQ(CMAKE_SOURCE_DIR "/libertine-data/upstart/application-legacy-single-.log", legacy_single);
 	g_free(legacy_single);
 
 	gchar * legacy_multiple = ubuntu_app_launch_application_log_path("bar");
-	EXPECT_STREQ(CMAKE_SOURCE_DIR "/upstart/application-legacy-bar-2342345.log", legacy_multiple);
+	EXPECT_STREQ(CMAKE_SOURCE_DIR "/libertine-data/upstart/application-legacy-bar-2342345.log", legacy_multiple);
 	g_free(legacy_multiple);
 }
 
@@ -503,6 +534,12 @@ TEST_F(LibUAL, ApplicationId)
 	EXPECT_EQ(nullptr, ubuntu_app_launch_triplet_to_app_id("com.test.no-json", NULL, NULL));
 	EXPECT_EQ(nullptr, ubuntu_app_launch_triplet_to_app_id("com.test.no-object", NULL, NULL));
 	EXPECT_EQ(nullptr, ubuntu_app_launch_triplet_to_app_id("com.test.no-version", NULL, NULL));
+
+	/* Libertine tests */
+	EXPECT_EQ(nullptr, ubuntu_app_launch_triplet_to_app_id("container-name", NULL, NULL));
+	EXPECT_EQ(nullptr, ubuntu_app_launch_triplet_to_app_id("container-name", "not-exist", NULL));
+	EXPECT_STREQ("container-name_test_0.0", ubuntu_app_launch_triplet_to_app_id("container-name", "test", NULL));
+	EXPECT_STREQ("container-name_user-app_0.0", ubuntu_app_launch_triplet_to_app_id("container-name", "user-app", NULL));
 }
 
 TEST_F(LibUAL, AppIdParse)
@@ -1446,3 +1483,189 @@ TEST_F(LibUAL, PauseResume)
 	
 	g_free(oomadjfile);
 }
+
+TEST_F(LibUAL, StartSessionHelper)
+{
+	DbusTestDbusMockObject * obj = dbus_test_dbus_mock_get_object(mock, "/com/test/untrusted/helper", "com.ubuntu.Upstart0_6.Job", NULL);
+	MirConnection * conn = mir_connect_sync("libual-test", "start-session-helper"); // Mocked, doesn't need cleaning up
+	MirPromptSession * msession = mir_connection_create_prompt_session_sync(conn, 5, nullptr, nullptr);
+
+	/* Building a temporary file and making an FD for it */
+	const char * filedata = "This is some data that we should get on the other side\n";
+	ASSERT_TRUE(g_file_set_contents(SESSION_TEMP_FILE, filedata, strlen(filedata), nullptr) == TRUE);
+	int mirfd = open(SESSION_TEMP_FILE, 0);
+	mir_mock_set_trusted_fd(mirfd);
+
+	/* Basic make sure we can send the event */
+	gchar * instance_id = ubuntu_app_launch_start_session_helper("untrusted-type", msession, "com.test.multiple_first_1.2.3", NULL);
+	ASSERT_NE(nullptr, instance_id);
+
+	guint len = 0;
+	const DbusTestDbusMockCall * calls = dbus_test_dbus_mock_object_get_method_calls(mock, obj, "Start", &len, NULL);
+	EXPECT_NE(nullptr, calls);
+	EXPECT_EQ(1, len);
+
+	EXPECT_STREQ("Start", calls->name);
+	EXPECT_EQ(2, g_variant_n_children(calls->params));
+
+	GVariant * block = g_variant_get_child_value(calls->params, 1);
+	EXPECT_TRUE(g_variant_get_boolean(block));
+	g_variant_unref(block);
+
+	/* Check the environment */
+	GVariant * env = g_variant_get_child_value(calls->params, 0);
+	EXPECT_TRUE(check_env(env, "APP_ID", "com.test.multiple_first_1.2.3"));
+	EXPECT_TRUE(check_env(env, "HELPER_TYPE", "untrusted-type"));
+	EXPECT_TRUE(check_env(env, "INSTANCE_ID", instance_id));
+
+	GVariant * mnamev = find_env(env, "UBUNTU_APP_LAUNCH_DEMANGLE_NAME");
+	ASSERT_NE(nullptr, mnamev); /* Have to assert because, eh, GVariant */
+	EXPECT_STREQ(g_dbus_connection_get_unique_name(bus), g_variant_get_string(mnamev, nullptr) + strlen("UBUNTU_APP_LAUNCH_DEMANGLE_NAME="));
+	GVariant * mpathv = find_env(env, "UBUNTU_APP_LAUNCH_DEMANGLE_PATH");
+	ASSERT_NE(nullptr, mpathv); /* Have to assert because, eh, GVariant */
+
+	g_variant_unref(env);
+
+	/* Setup environment for call */
+	const gchar * mname = g_variant_get_string(mnamev, nullptr);
+	mname += strlen("UBUNTU_APP_LAUNCH_DEMANGLE_NAME=");
+	g_setenv("UBUNTU_APP_LAUNCH_DEMANGLE_NAME", mname, TRUE);
+	g_variant_unref(mnamev);
+
+	const gchar * mpath = g_variant_get_string(mpathv, nullptr);
+	mpath += strlen("UBUNTU_APP_LAUNCH_DEMANGLE_PATH=");
+	g_setenv("UBUNTU_APP_LAUNCH_DEMANGLE_PATH", mpath, TRUE);
+	g_variant_unref(mpathv);
+
+	/* Exec our tool */
+	std::promise<std::string> outputpromise;
+	std::thread t([&outputpromise]() {
+		gchar * socketstdout = nullptr;
+		GError * error = nullptr;
+		g_unsetenv("G_MESSAGES_DEBUG");
+
+		g_spawn_command_line_sync(
+				SOCKET_DEMANGLER " " SOCKET_TOOL,
+				&socketstdout,
+				nullptr,
+				nullptr,
+				&error);
+
+		if (error != nullptr) {
+			fprintf(stderr, "Unable to spawn '" SOCKET_DEMANGLER " " SOCKET_TOOL "': %s\n", error->message);
+			g_error_free(error);
+			outputpromise.set_value(std::string(""));
+		} else {
+			outputpromise.set_value(std::string(socketstdout));
+			g_free(socketstdout);
+		}
+	});
+	t.detach();
+
+	auto outputfuture = outputpromise.get_future();
+	while (outputfuture.wait_for(std::chrono::milliseconds{1}) != std::future_status::ready) {
+		pause();
+	}
+
+	ASSERT_STREQ(filedata, outputfuture.get().c_str());
+
+	ASSERT_TRUE(dbus_test_dbus_mock_object_clear_method_calls(mock, obj, NULL));
+
+	return;
+}
+
+TEST_F(LibUAL, SetExec)
+{
+	DbusTestDbusMockObject * obj = dbus_test_dbus_mock_get_object(mock, "/com/ubuntu/Upstart", "com.ubuntu.Upstart0_6", NULL);
+
+	const char * exec = "lets exec this";
+
+	g_setenv("UPSTART_JOB", "fubar", TRUE);
+	g_unsetenv("UBUNTU_APP_LAUNCH_DEMANGLE_NAME");
+	EXPECT_TRUE(ubuntu_app_launch_helper_set_exec(exec, NULL));
+
+	guint len = 0;
+	const DbusTestDbusMockCall * calls = dbus_test_dbus_mock_object_get_method_calls(mock, obj, "SetEnv", &len, NULL);
+	ASSERT_NE(nullptr, calls);
+	EXPECT_EQ(1, len);
+
+	gchar * appexecstr = g_strdup_printf("APP_EXEC=%s", exec);
+	GVariant * appexecenv = g_variant_get_child_value(calls[0].params, 1);
+	EXPECT_STREQ(appexecstr, g_variant_get_string(appexecenv, nullptr));
+	g_variant_unref(appexecenv);
+	g_free(appexecstr);
+
+	ASSERT_TRUE(dbus_test_dbus_mock_object_clear_method_calls(mock, obj, NULL));
+
+	/* Now check for the demangler */
+	g_setenv("UBUNTU_APP_LAUNCH_DEMANGLE_NAME", g_dbus_connection_get_unique_name(bus), TRUE);
+	EXPECT_TRUE(ubuntu_app_launch_helper_set_exec(exec, NULL));
+
+	calls = dbus_test_dbus_mock_object_get_method_calls(mock, obj, "SetEnv", &len, NULL);
+	ASSERT_NE(nullptr, calls);
+	EXPECT_EQ(1, len);
+
+	gchar * demangleexecstr = g_strdup_printf("APP_EXEC=%s %s", SOCKET_DEMANGLER_INSTALL, exec);
+	appexecenv = g_variant_get_child_value(calls[0].params, 1);
+	EXPECT_STREQ(demangleexecstr, g_variant_get_string(appexecenv, nullptr));
+	g_variant_unref(appexecenv);
+	g_free(demangleexecstr);
+
+	ASSERT_TRUE(dbus_test_dbus_mock_object_clear_method_calls(mock, obj, NULL));
+
+	/* Now check for the directory */
+	g_setenv("UBUNTU_APP_LAUNCH_DEMANGLE_NAME", g_dbus_connection_get_unique_name(bus), TRUE);
+	EXPECT_TRUE(ubuntu_app_launch_helper_set_exec(exec, "/not/a/real/directory"));
+
+	calls = dbus_test_dbus_mock_object_get_method_calls(mock, obj, "SetEnv", &len, NULL);
+	ASSERT_NE(nullptr, calls);
+	EXPECT_EQ(2, len);
+
+	appexecenv = g_variant_get_child_value(calls[1].params, 1);
+	EXPECT_STREQ("APP_DIR=/not/a/real/directory", g_variant_get_string(appexecenv, nullptr));
+	g_variant_unref(appexecenv);
+
+	ASSERT_TRUE(dbus_test_dbus_mock_object_clear_method_calls(mock, obj, NULL));
+}
+
+TEST_F(LibUAL, AppInfo)
+{
+	g_setenv("TEST_CLICK_DB", "click-db-dir", TRUE);
+	g_setenv("TEST_CLICK_USER", "test-user", TRUE);
+
+	char * dir = nullptr;
+	char * file = nullptr;
+
+	/* Basics */
+	EXPECT_TRUE(ubuntu_app_launch_application_info("com.test.good_application_1.2.3", nullptr, nullptr));
+	EXPECT_FALSE(ubuntu_app_launch_application_info("com.test.bad_not-app_1.3.3.7", nullptr, nullptr));
+	EXPECT_FALSE(ubuntu_app_launch_application_info(nullptr, nullptr, nullptr));
+
+	/* Ensure false doesn't set the values */
+	EXPECT_FALSE(ubuntu_app_launch_application_info("com.test.bad_not-app_1.3.3.7", &dir, &file));
+	EXPECT_EQ(nullptr, dir);
+	EXPECT_EQ(nullptr, file);
+	g_clear_pointer(&dir, g_free);
+	g_clear_pointer(&file, g_free);
+
+	/* Correct values from a click */
+	EXPECT_TRUE(ubuntu_app_launch_application_info("com.test.good_application_1.2.3", &dir, &file));
+	EXPECT_STREQ(CMAKE_SOURCE_DIR "/click-root-dir/.click/users/test-user/com.test.good", dir);
+	EXPECT_STREQ("application.desktop", file);
+	g_clear_pointer(&dir, g_free);
+	g_clear_pointer(&file, g_free);
+
+	/* Correct values from a legacy */
+	EXPECT_TRUE(ubuntu_app_launch_application_info("bar", &dir, &file));
+	EXPECT_STREQ(CMAKE_SOURCE_DIR, dir);
+	EXPECT_STREQ("applications/bar.desktop", file);
+	g_clear_pointer(&dir, g_free);
+	g_clear_pointer(&file, g_free);
+
+	/* Correct values for libertine */
+	EXPECT_TRUE(ubuntu_app_launch_application_info("container-name_test_0.0", &dir, &file));
+	EXPECT_STREQ(CMAKE_SOURCE_DIR "/libertine-data/libertine-container/container-name/rootfs/usr/share", dir);
+	EXPECT_STREQ("applications/test.desktop", file);
+	g_clear_pointer(&dir, g_free);
+	g_clear_pointer(&file, g_free);
+}
